@@ -6,7 +6,7 @@ use strict;
 use warnings;
 use v5.10;    # needed for state variable
 
-our $VERSION = '0.100'; # VERSION
+our $VERSION = '0.101'; # VERSION
 our $AUTHORITY = 'cpan:NIGELM'; # AUTHORITY
 
 use Moose;
@@ -373,7 +373,7 @@ method _build_returned_objects ($response) {
             $self->_debug("API: building a set of [$thing_type] objects") if ( $self->debug );
             foreach my $thing ( $self->_listify( $hash->{$thing_type} ) ) {
                 my $object = VMware::vCloudDirector2::Object->new(
-                    hash            => { $thing_type => $thing },
+                    hash            => $thing,
                     api             => $self,
                     _partial_object => 1
                 );
@@ -486,7 +486,7 @@ VMware::vCloudDirector2::API - Module to do stuff!
 
 =head1 VERSION
 
-version 0.100
+version 0.101
 
 =head2 Attributes
 
