@@ -5,14 +5,16 @@ package VMware::vCloudDirector2::Object;
 use strict;
 use warnings;
 
-our $VERSION = '0.100'; # VERSION
+our $VERSION = '0.101'; # VERSION
 our $AUTHORITY = 'cpan:NIGELM'; # AUTHORITY
 
 use Moose;
 use Method::Signatures;
-use Ref::Util qw(is_plain_hashref);
-use Lingua::EN::Inflexion;
-use VMware::vCloudDirector2::ObjectContent;
+use MooseX::Types::URI qw(Uri);
+use Const::Fast;
+use Ref::Util qw(is_plain_hashref is_plain_arrayref);
+use VMware::vCloudDirector2::Link;
+use VMware::vCloudDirector2::Error;
 
 # ------------------------------------------------------------------------
 
@@ -25,30 +27,141 @@ has api => (
     documentation => 'API we use'
 );
 
-has content => (
-    is            => 'ro',
-    isa           => 'VMware::vCloudDirector2::ObjectContent',
-    predicate     => 'has_content',
-    writer        => '_set_content',
-    documentation => 'The underlying content object',
-    handles       => [qw( mime_type href type name )],
-);
+has mime_type => ( is => 'ro', isa => 'Str', required => 1 );
+has href => ( is => 'ro', isa => Uri,   required => 1, coerce => 1 );
+has type => ( is => 'ro', isa => 'Str', required => 1 );
+has uuid => ( is => 'ro', isa => 'Str', builder  => '_build_uuid', lazy => 1 );
+has name =>
+    ( is => 'ro', isa => 'Str', predicate => 'has_name', lazy => 1, builder => '_build_name' );
+has id => ( is => 'ro', isa => 'Str', predicate => 'has_id', lazy => 1, builder => '_build_id' );
 
 has _partial_object => ( is => 'rw', isa => 'Bool', default => 0 );
-
-# delegates that force a full object to be pulled
-method hash () { return $self->inflate->content->hash; }
-method links () { return $self->inflate->content->links; }
-method id () { return $self->inflate->content->id; }
-method uuid () { return ( split( /\//, $self->href ) )[-1]; }
+has is_json         => ( is => 'rw', isa => 'Bool', default => 0 );
 
 # ------------------------------------------------------------------------
-method BUILD ($args) {
+around BUILDARGS => sub {
+    my ( $orig, $class, $first, @rest ) = @_;
 
-    $self->_set_content(
-        VMware::vCloudDirector2::ObjectContent->new( object => $self, hash => $args->{hash} ) );
+    my $params = is_plain_hashref($first) ? $first : { $first, @rest };
+    if ( $params->{hash} ) {
+        my $hash = $params->{hash};
+
+        # copy elements into object attributes
+        foreach (qw[href name id]) {
+            $params->{$_} = $hash->{$_} if ( exists( $hash->{$_} ) and defined( $hash->{$_} ) );
+        }
+
+        # set the object type and mime_type
+        if ( exists( $hash->{type} ) ) {
+            $params->{mime_type} = $hash->{type};
+            $params->{type}      = $1
+                if ( $hash->{type} =~ m!^application/vnd\..*\.(\w+)\+(json|xml)$! );
+            $params->{is_json} = ( $2 eq 'json' ) ? 1 : 0;
+        }
+
+        # if this has a links section it is a complete object, otherwise its partial
+        if ( exists( $hash->{link} ) ) {
+            $params->{_partial_object} = 0;
+            const $params->{hash} => $hash;    # force hash read-only to stop people playing
+        }
+        else {
+            $params->{_partial_object} = 1;
+            delete( $params->{hash} );         # do not populate the hash in the partial object
+        }
+    }
+    else {
+        # no hash so this must be a partial object
+        $params->{_partial_object} = 1;
+    }
+    return $class->$orig($params);
+};
+
+# ------------------------------------------------------------------------
+has hash => (
+    is      => 'ro',
+    traits  => ['Hash'],
+    isa     => 'HashRef',
+    builder => '_build_hash',
+    clearer => '_clear_hash',
+    lazy    => 1,
+    handles => { get_hash_item => 'get', exists_hash_item => 'exists', }
+);
+
+method _build_hash () {
+
+    # fetch object content
+    const my $hash => $self->api->GET_hash( $self->href );
+    $self->api->_debug(
+        sprintf(
+            'Object: %s a [%s]',
+            ( $self->_partial_object ? 'Inflated' : 'Refetched' ),
+            $self->type
+        )
+    ) if ( $self->api->debug );
+
+    # mark as being a whole object
+    $self->_partial_object(0);
+
+    return $hash;
 }
 
+method _build_name () { return $self->get_hash_item('name'); }
+method _build_id () { return $self->get_hash_item('id'); }
+
+method _build_uuid () {
+
+    # The UUID is in the href - return the first match
+    my $path = lc( $self->href->path() );
+    return $1
+        if ( $path =~ m|\b([0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{12})\b| );
+    return;
+}
+
+# ------------------------------------------------------------------------
+
+has _links => (
+    is      => 'ro',
+    traits  => ['Array'],
+    isa     => 'ArrayRef[VMware::vCloudDirector2::Link]',
+    lazy    => 1,
+    builder => '_build_links',
+    handles => { links => 'elements', },
+);
+has _all_links => (
+    is      => 'ro',
+    traits  => ['Array'],
+    isa     => 'ArrayRef[VMware::vCloudDirector2::Link]',
+    lazy    => 1,
+    builder => '_build_all_links',
+    handles => { all_links => 'elements', },
+);
+
+method _build_links () {
+    my @links = grep { $_->is_json } $self->all_links;
+    return \@links;
+}
+
+method _build_all_links () {
+    my @links;
+    if ( exists( $self->hash->{link} ) ) {
+        push( @links, VMware::vCloudDirector2::Link->new( hash => $_, object => $self ) )
+            foreach ( $self->_listify( $self->hash->{link} ) );
+    }
+    return \@links;
+}
+
+# ------------------------------------------------------------------------
+
+
+has is_admin_object => (
+    is            => 'ro',
+    isa           => 'Bool',
+    lazy          => 1,
+    builder       => '_build_is_admin_object',
+    documentation => 'Is this an admin level object?',
+);
+method _build_is_admin_object () { return ( $self->href->path() =~ m|/api/admin/| ) ? 1 : 0; }
+
 # ------------------------------------------------------------------------
 
 
@@ -59,17 +172,12 @@ method inflate () {
 
 # ------------------------------------------------------------------------
 method refetch () {
-    my $hash = $self->api->GET_hash( $self->href );
-    $self->_set_content(
-        VMware::vCloudDirector2::ObjectContent->new( object => $self, hash => $hash ) );
-    $self->api->_debug(
-        sprintf(
-            'Object: %s a [%s]',
-            ( $self->_partial_object ? 'Inflated' : 'Refetched' ),
-            $self->type
-        )
-    ) if ( $self->api->debug );
-    $self->_partial_object(0);
+
+    # simplest way to force the object to be refetched is to clear the hash
+    # and then request it which forces a lazy eval
+    $self->_clear_hash;
+    $self->hash;
+
     return $self;
 }
 
@@ -78,8 +186,7 @@ method refetch () {
 
 method find_links (:$name, :$type, :$rel) {
     my @matched_links;
-    my $links = $self->links;
-    foreach my $link ( @{$links} ) {
+    foreach my $link ( $self->links ) {
         if ( not( defined($rel) ) or ( $rel eq ( $link->rel || '' ) ) ) {
             if ( not( defined($type) ) or ( $type eq ( $link->type || '' ) ) ) {
                 if ( not( defined($name) ) or ( $name eq ( $link->name || '' ) ) ) {
@@ -94,6 +201,21 @@ method find_links (:$name, :$type, :$rel) {
 # ------------------------------------------------------------------------
 
 
+method find_link (@criteria) {
+    my @matched_links = $self->find_links(@criteria);
+    unless ( scalar(@matched_links) ) {
+        VMware::vCloudDirector2::Error->throw(
+            {   message => sprintf( "No links matching criteria: %s", join( ', ', @criteria ) ),
+                object  => $self
+            }
+        );
+    }
+    return $matched_links[0];
+}
+
+# ------------------------------------------------------------------------
+
+
 method fetch_links (@search_items) {
     my @matched_objects;
     foreach my $link ( $self->find_links(@search_items) ) {
@@ -108,9 +230,9 @@ method _create_object ($hash, $type='Thing') {
     # if thing has Link content within it then it is a full object, otherwise it
     # is just a stub
     my $object = VMware::vCloudDirector2::Object->new(
-        hash            => { $type => $hash },
+        hash            => $hash,
         api             => $self->api,
-        _partial_object => ( exists( $hash->{Link} ) ) ? 0 : 1,
+        _partial_object => ( exists( $hash->{link} ) ) ? 0 : 1,
     );
     $self->api->_debug(
         sprintf(
@@ -127,11 +249,9 @@ method _create_object ($hash, $type='Thing') {
 
 method build_sub_objects ($type) {
     my @objects;
-    my $container_type = noun($type)->plural;
-    return
-        unless ( exists( $self->hash->{$container_type} )
-        and is_plain_hashref( $self->hash->{$container_type} ) );
-    foreach my $thing ( $self->_listify( $self->hash->{$container_type}{$type} ) ) {
+
+    return unless ( exists( $self->hash->{$type} ) );
+    foreach my $thing ( $self->_listify( $self->hash->{$type} ) ) {
         push( @objects, $self->_create_object( $thing, $type ) );
     }
     return @objects;
@@ -139,10 +259,10 @@ method build_sub_objects ($type) {
 
 method build_children_objects () {
     my $hash = $self->hash;
-    return unless ( exists( $hash->{Children} ) and is_plain_hashref( $hash->{Children} ) );
+    return unless ( exists( $hash->{children} ) and is_plain_hashref( $hash->{children} ) );
     my @objects;
-    foreach my $key ( keys %{ $hash->{Children} } ) {
-        foreach my $thing ( $self->_listify( $self->hash->{Children}{$key} ) ) {
+    foreach my $key ( keys %{ $hash->{children} } ) {
+        foreach my $thing ( $self->_listify( $self->hash->{children}{$key} ) ) {
             push( @objects, $self->_create_object( $thing, $key ) );
         }
     }
@@ -163,6 +283,20 @@ method POST ($hash) { return $self->api->POST( $self->href, $hash, $self->mime_t
 
 method PUT ($hash) { return $self->api->PUT( $self->href, $hash, $self->mime_type ); }
 
+# ------------------------------------------------------------------------
+
+
+method fetch_admin_object ($subpath?) {
+    if ( $self->is_admin_object ) {
+        return $self;
+    }
+    else {
+        my $path = join( '/', '/api/admin', $self->type, $self->uuid );
+        $path .= '/' . $subpath if ( defined($subpath) );
+        return $self->api->GET($path);
+    }
+}
+
 # ------------------------------------------------------------------------
 method _listify ($thing) { !defined $thing ? () : ( ( ref $thing eq 'ARRAY' ) ? @{$thing} : $thing ) }
 
@@ -184,7 +318,7 @@ VMware::vCloudDirector2::Object - Module to contain an object!
 
 =head1 VERSION
 
-version 0.100
+version 0.101
 
 =head2 Attributes
 
@@ -204,13 +338,24 @@ A reference to the hash returned from the vCloud API.  Forces object inflation.
 
 =head3 links
 
-An array references to the links contained in this object.  Forces object
-inflation.
+Returns L<VMware::vCloudDirector2::Link> objects for each of the JSON targetted
+links contained in this object.  Forces object inflation.
+
+=head3 all_links
+
+Returns L<VMware::vCloudDirector2::Link> objects for each of the links
+contained in this object.  Will typically return two links per thing - one to
+the XML version, one to the JSON version.  Forces object inflation.
 
 =head3 id
 
 The id attribute from the returned vCloud JSON.  Forces object inflation.
 
+=head3 is_admin_object
+
+This determines, based on the href path, whether or not this is an admin
+object.
+
 =head2 Methods
 
 =head3 inflate
@@ -246,6 +391,15 @@ The rel of the link
 
 The return value is a list of link objects.
 
+=head3 find_link
+
+Finds and returns one link that matches the search criteria, exactly as
+L<find_links>, except that if no links are found an exception is thrown.  If
+multiple links match then the first one returned (normally the first one back
+from the API) would be returned.
+
+The return value is a single link object.
+
 =head3 fetch_links
 
 As per L</find_links> except that each link found is fetched and expanded up as
@@ -257,10 +411,6 @@ Given a type (specifically a key used within the current object hash), grabs
 the descendants of that key and instantiates them as partial objects (they can
 then be inflated into full objects).
 
-Due to the structure of the XML there will always be two layers, the inner
-named singular thing, and the outer named as the plural of thing.  Hence this
-does magic with the language inflection module.
-
 =head3 DELETE
 
 Make a delete request to the URL in this link.  Returns Objects.  Failure will
@@ -283,6 +433,16 @@ Make a put request with the specified payload to the URL in this link.  Returns
 Objects.  Failure will generate an exception.  See
 L<VMware::vCloudDirector2::API/PUT>.
 
+=head3 fetch_admin_object
+
+If this is already an admin object (ie C<is_admin_object> is true), then this
+object is returned.
+
+Otherwise, the path is modified to point to the admin API object and the object
+is fetched.  Since this only exists for a subset of objects there is a
+reasonable chance that just attempting this will lead to an exception  being
+thrown due to a non-existant object being requested.
+
 =head1 AUTHOR
 
 Nigel Metheringham <nigelm@cpan.org>
